#!/usr/bin/env python3

import os
import threading
import traceback
import datetime
import time
from flask import Flask, request, jsonify, abort
import json
from socket import *
import select
from lib import kms_core
from lib.db import (
    get_db, set_db_folder, close_db,
    set_db_uri, ActivateMode,
    format_serial_data, format_client_data
)
import logging
import logging.handlers
from functools import wraps
import lib.contrib.paho.mqtt.client as mqtt
from lib.contrib.paho.mqtt import properties
import configparser
import urllib.request
import urllib.parse

srcdir = os.path.dirname(__file__)


class KMSConfig(configparser.ConfigParser):
    current_files = None

    def read(self, filenames, encoding=None):
        self.current_files = filenames
        return super().read(filenames, encoding)

    def reload(self):
        try:
            self.read(self.current_files)
        except Exception:
            app.logger.exception('failed reload config')


# get conf
DEFAULT_CONF = os.path.join(srcdir, 'conf/app_template.conf')
CUSTOM_CONF = os.path.join(srcdir, 'conf/app.conf')
app_conf = KMSConfig()
app_conf.read([DEFAULT_CONF, CUSTOM_CONF], encoding="utf-8")

# get VERSION
for version_file in (os.path.join(srcdir, 'VERSION'), os.path.join(srcdir, '../VERSION')):
    if os.path.exists(version_file):
        with open(version_file) as f:
            __version__ = f.read().strip()
        break

app = Flask('kylin-kms', instance_path='/var/lib/kylin-kms-daemon')
app.logger.setLevel(logging.DEBUG)
# 不同系统中默认handler不一样
app.logger.handlers = []
formatter = logging.Formatter(
    '[%(asctime)s] %(levelname)s in %(module)s: %(message)s')
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
app.logger.addHandler(stream_handler)
file_handler = logging.handlers.RotatingFileHandler(filename='/var/log/kylin-kms-daemon.log',
                                                    maxBytes=1024*1024, backupCount=5)
file_handler.setFormatter(formatter)
app.logger.addHandler(file_handler)
set_db_folder(os.path.join(app.instance_path, 'database'))
set_db_uri(app_conf.get('server', 'db_uri'))
kms_core.logger = app.logger
kms_core.init_manager(app.instance_path)
kms_core.instance_path = app.instance_path
app.teardown_appcontext(close_db)

app.logger.info('Kylin KMS Daemon Starting')
app.logger.info('VERSION: %s', __version__)
app.logger.info('KMS ID: %s', kms_core.manager.mac)

client = mqtt.Client(protocol=mqtt.MQTTv5)
publish_properties = mqtt.Properties(properties.PacketTypes.PUBLISH)
publish_properties.MessageExpiryInterval = 60
protocol_version = '2.2'

try:
    os.makedirs(app.instance_path)
except OSError:
    pass


def require_mode(mode_name):
    """
    检查模式是否与配置文件中的配置项匹配
    @param mode_name: 设定的模式名称
    @return:
    """
    def wrapper(func):
        @wraps(func)
        def mode_correct(*args, **kwargs):
            if app_conf.get('server', 'mode') != mode_name:
                return 'Server mode is not match!'
            return func(*args, **kwargs)
        return mode_correct
    return wrapper


def is_local():
    if request.remote_addr.startswith('127.0.0.'):
        return True
    if request.remote_addr == request.host.split(':')[0]:
        return True
    # whitelist
    white_list_file = os.path.join(app.instance_path, 'whitelist')
    if os.path.isfile(white_list_file):
        try:
            with open(white_list_file) as f:
                for line in f.read().strip().split('\n'):
                    if request.remote_addr == line.strip():
                        return True
        except:
            pass
    return False


def requires_local(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if not is_local():
            abort(403, 'Private Function')
        return f(*args, **kwargs)
    return decorated_function


@app.route('/')
def hello():
    print(dir(request))
    print(request.remote_addr)
    print(request.host)
    return 'Hello this is Kylin KMS daemon\n'


@app.route('/api/server/import-master', methods=['POST'])
@require_mode('slave')
def backup_databases():
    data = request.stream.read()
    kms_core.import_master.start_thread(data)
    return jsonify({'status': 'success'})


@app.route('/api/server/version')
def get_all_serial():
    db = get_db()
    all = request.args.get('all', '')

    kms_core.update_type2_status()

    if all == 'yes':
        serials = db().select(db.serial.ALL)
    else:
        # 可激活数量为0，且没有激活记录（不计算已取消的）的license，将其隐藏
        serial_rows = db(
            ~((db.serial.f_type != 2) &
              (db.serial.f_amount == 0) &
              (db.serial.f_used == 0) &
              (db.serial.f_mode < ActivateMode.unlimited)
              )
        ).select(db.serial.ALL)
        # 隐藏已激活数量为0的厂商授权衍生license
        serials = [row for row in serial_rows if
                   not (row.f_type == 2 and row.f_used == 0)]

    return jsonify({
        'return_code': 0,
        'return_msg': 'success',
        'payload': [format_serial_data(row) for row in serials],
        'server': {'version': __version__,
                   'mac': kms_core.manager.mac,
                   'passive': kms_core.manager.passive
                   }
    })


@app.route('/api/server/clientlist')
def get_activated():
    try:
        limit = int(request.args.get('limit'))
    except:
        limit = 0
    try:
        page = int(request.args.get('offset'))
    except:
        page = 0
    try:
        today = int(request.args.get('today')) == 0
    except:
        today = False
    try:
        since = int(request.args.get('since'))
    except:
        since = 0

    serial = request.args.get('serial')

    db = get_db()

    query = (db.activation.f_serial == db.serial.f_serial)

    if today:
        today_start = time.mktime(datetime.date.today().timetuple())
        query &= (db.activation.f_change >= today_start)

    if since:
        query &= (db.activation.f_change >= since - 5)

    if serial:
        query &= (db.serial.f_serial == serial)

    if limit:
        rows = db(query).select(orderby=db.activation.id,
                                limitby=(limit * page, limit * (page + 1)))
    else:
        rows = db(query).select(orderby=db.activation.id)
    total = db(query).count(db.activation.id)

    # 取rows中的regcode，并获取每个regcode的client_extend_attr表中数据信息
    regcode_list = [row.activation.f_regcode for row in rows]
    extend_attr_rows = db(
            (db.client_extend_attr.f_regcode.belongs(regcode_list))
    ).select()
    # 去重，重复数据可能是由于单个client同时发送了多个激活请求
    clients = set()
    client_datas = []
    for row in rows:
        if row.activation.f_regcode in clients:
            # TODO: remove duplicate row
            continue
        client_datas.append(format_client_data(row, extend_attr_rows))
        clients.add(row.activation.f_regcode)

    return jsonify({
        'payload': client_datas,
        'return_code': 0,
        'return_msg': '调用成功',
        'total': total,
        'server': {'version': __version__,
                   'mac': kms_core.manager.mac,
                   'mode': app_conf.get('server', 'mode')}
    })


@app.route('/api/server/clientinfo')
def get_client():
    regcode = request.args.get('client_Key')
    if not regcode:
        return jsonify({'return_code': 1, 'return_msg': '请提供client_Key参数'})
    if len(regcode) == 20:
        regcode = kms_core.code_add_hyphen(regcode)
    if len(regcode) != 24:
        return jsonify({'return_code': 2, 'return_msg': '注册码长度错误'})

    # find activation
    db = get_db()
    row = db(
        (db.activation.f_regcode == regcode) &
        (db.activation.f_serial == db.serial.f_serial)
    ).select().first()
    if not row:
        return jsonify({'return_code': 3, 'return_msg': '未找到指定终端'})

    # 获取client_extend_attr表中的信息
    extend_attr_rows = db(
            (db.client_extend_attr.f_regcode == row.activation.f_regcode)
    ).select()

    return jsonify({
        'return_code': 0,
        'return_msg': '调用成功',
        'payload': format_client_data(row, extend_attr_rows)
    })


@app.route('/api/server/cert', methods=['POST'])
def import_license():
    license_data = request.stream.read()
    try:
        data_dict = json.loads(license_data.decode('utf-8'))
        license_data = data_dict['cert_info']
    except:
        pass
    try:
        kms_core.manager.import_license(license_data)
        ret_code = 0
        msg = '导入成功'
    except Exception as e:
        ret_code = 1
        msg = str(e)

    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/enablelicense', methods=['POST'])
@requires_local
def enable_license():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    serial = data_dict['serial']

    db = get_db()
    row = db(db.serial.f_serial == serial).select().first()
    if row:
        if not row.f_enabled:
            # enable this
            row.update_record(f_enabled=1)
            db.commit()
        ret_code = 0
        msg = 'Enabled'
        kms_core.manager.put_data_2_queue(serial)
    else:
        ret_code = 1
        msg = 'Serial Not Found'

    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/disablelicense', methods=['POST'])
@requires_local
def disable_license():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    serial = data_dict['serial']

    db = get_db()
    count = db(db.serial.f_serial == serial).update(f_enabled=0)
    if count:
        db.commit()
        ret_code = 0
        msg = 'Disabled'
        kms_core.manager.put_data_2_queue(serial)
    else:
        ret_code = 1
        msg = 'Serial Not Found'

    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/set_validity', methods=['POST'])
@requires_local
def set_validity():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    # example: {'serial':'100000', 'validity':30}
    serial = data_dict['serial']
    validity = data_dict['validity']

    ret_code, msg = kms_core.set_serial_validity(serial, validity)

    kms_core.manager.put_data_2_queue(serial)
    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/editlicense', methods=['POST'])
@requires_local
def edit_license():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    serial = data_dict['serial']
    # 一次只能修改一种属性
    fields = set(data_dict.keys()) & set(
        ['validity', 'enabled', 'deactivateconfirm', 'activate_for', 'match_store'])
    if len(fields) != 1:
        ret_code = 1
        msg = 'You can only update one field'
    else:
        field = fields.pop()
        value = data_dict[field]
        db = get_db()
        if field == 'enabled':
            if value:
                # enable
                row = db(db.serial.f_serial == serial).select().first()
                if row:
                    if not row.f_enabled:
                        # enable this
                        row.update_record(f_enabled=1)
                        db.commit()
                    ret_code = 0
                    msg = 'Enabled'
                else:
                    ret_code = 1
                    msg = 'Serial Not Found'
            else:
                # disable
                count = db(db.serial.f_serial == serial).update(f_enabled=0)
                if count:
                    db.commit()
                    ret_code = 0
                    msg = 'Disabled'
                else:
                    ret_code = 1
                    msg = 'Serial Not Found'
        elif field == 'validity':
            ret_code, msg = kms_core.set_serial_validity(serial, value)
        elif field == 'deactivateconfirm':
            value = int(bool(value))
            count = db(db.serial.f_serial == serial).update(
                f_deactivate_confirm=value)
            if count:
                db.commit()
                ret_code = 0
                msg = 'Success'
            else:
                ret_code = 1
                msg = 'Serial Not Found'
        elif field == 'activate_for':
            ret_code, msg = kms_core.set_general_activate(serial, value)
        elif field == 'match_store':
            value = int(bool(value))
            count = db(db.serial.f_serial == serial).update(
                f_match_store=value)
            if count:
                db.commit()
                ret_code = 0
                msg = 'Success'
            else:
                ret_code = 1
                msg = 'Serial Not Found'

        kms_core.manager.put_data_2_queue(serial)
    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/deactivate', methods=['POST'])
@requires_local
def deactivate():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    # example: {'clientKey':'AAAA-BBBB-CCCC-DDDD-EEEE', 'reason': 'test'}
    regcode = data_dict.get('clientKey') or data_dict.get('clientkey')
    reason = data_dict.get('reason')
    if not regcode:
        ret_code = 1
        msg = 'need a clientKey'
    else:
        regcode = regcode.replace('-', '')
        db = get_db()
        row = db(
            (db.activation.f_regcode == kms_core.code_add_hyphen(regcode)) &
            (db.activation.f_serial == db.serial.f_serial)
        ).select().first()
        if not row:
            ret_code = 2
            msg = 'no such client'
        else:
            status, msg = kms_core.deactivate(
                row.serial.f_serial, regcode, reason, local=True)
            if status == 'SUCCESS':
                ret_code = 0
            else:
                ret_code = 3

    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/confirm_deactivate', methods=['POST'])
@requires_local
def confirm_deactivate():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    # example: {'clientKey':'AAAA-BBBB-CCCC-DDDD-EEEE', 'allow': True}
    regcode = data_dict.get('clientKey') or data_dict.get('clientkey')
    allow = data_dict.get('allow')
    if not regcode:
        ret_code = 1
        msg = 'need a clientKey'
    else:
        regcode = regcode.replace('-', '')
        db = get_db()
        row = db(
            (db.activation.f_regcode == kms_core.code_add_hyphen(regcode)) &
            (db.activation.f_serial == db.serial.f_serial)
        ).select().first()
        if not row:
            ret_code = 2
            msg = 'no such client'
        else:
            if allow == True:
                status, msg = kms_core.deactivate(
                    row.serial.f_serial, regcode, '', local=True, confirm=True)
                if status == 'SUCCESS':
                    ret_code = 0
                else:
                    ret_code = 3
            else:
                if row.activation.f_status == 4:
                    ret_code = 4
                    msg = 'Client already deactivated'
                else:
                    row.activation.update_record(
                        f_deactivate_status=3, f_change=time.time())
                    ret_code = 0
                    msg = 'Deactivation rejected'

    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/dump', methods=['GET'])
def dump_license():
    return kms_core.manager.dump_all()


@app.route('/api/server/uploadtokylin')
def upload_to_kylin():
    return


@app.route('/api/server/batchset_storetag', methods=['POST'])
def batchset_storetag():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    # 本api只能修改match_store属性
    fields = set(data_dict.keys()) & set(['match_store'])
    if len(fields) != 1:
        ret_code = 1
        msg = 'You can only update match_store field'
    else:
        value = data_dict['match_store']
        db = get_db()
        value = int(bool(value))
        ret_code = 0
        msg = 'Success'
        count = db(db.serial.id > 0).update(
            f_match_store=value)
        if count:
            db.commit()

    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/feature', methods=['POST'])
def update_feature():
    data = request.stream.read()
    data_dict = json.loads(data.decode('utf-8'))
    # 暂时只能修改passive属性
    if 'passive' in data_dict:
        value = bool(data_dict['passive'])
        app_conf.set('server', 'passive_mode', str(value))
        kms_core.manager.passive = value
        with open(CUSTOM_CONF, 'w', encoding='utf-8') as f:
            app_conf.write(f)
        return jsonify({'return_code': 0, 'return_msg': 'Success'})
    else:
        return jsonify({'return_code': 1, 'return_msg': 'No supported feature'})


@app.route('/api/server/monitor')
def monitor():
    """
    获取KMS服务端状态
    @return: response
    """
    return jsonify(
        dict(
            code=200,
            status=1,
            msg='Server is alive!',
            mode=app_conf.get('server', 'mode')
        )
    )


@app.route('/api/server/promote')
def promote():
    """
    将服务端升级为master端
    @return: response
    """
    app_conf.set('server', 'mode', 'master')
    with open(CUSTOM_CONF, 'w', encoding='utf-8') as f:
        app_conf.write(f)
    return jsonify(
        dict(
            code=200,
            status=1,
            msg='Promote server to master!'
        )
    )


@app.route('/api/server/demote')
def demote():
    """
    将服务端降级为slave端
    @return: response
    """
    app_conf.set('server', 'mode', 'slave')
    with open(CUSTOM_CONF, 'w', encoding='utf-8') as f:
        app_conf.write(f)
    return jsonify(
        dict(
            code=200,
            status=1,
            msg='Demote server to slave!'
        )
    )


def load_bind_conf(option, default):
    section = 'ports'
    try:
        api_bind = app_conf.get(section, option)
        host, _port = api_bind.split(':')
        port = int(_port)
    except Exception:
        host, port = default[0], default[1]
        app.logger.error(
            '%s.%s configure error, fallback to %s:%d',
            section, option, host, port
        )
    return host, port


def listen_http():
    os.environ["WERKZEUG_RUN_MAIN"] = "true"
    host, port = load_bind_conf('api_bind', ('0.0.0.0', 8090))
    app.run(host=host, port=port)


@require_mode('master')
def dealrequest(request, cs, addr):
    app.logger.debug(request)
    try:
        data = request.split(':')
        command = data[0].strip()

        if command == 'KY-ACTIVATE':
            serial = data[1]
            regcode = data[2]
            assert command == 'KY-ACTIVATE'
            try:
                hwinfo = data[3]
            except IndexError:
                hwinfo = ''
            try:
                oem_serial = data[4]
            except IndexError:
                oem_serial = ''
            try:
                vendor = data[5]
            except IndexError:
                vendor = ''
            try:
                model = data[6]
            except IndexError:
                model = ''
        elif command == 'KY-DEACTIVATE':
            serial = data[1]
            regcode = data[2]
            try:
                reason = data[3]
            except IndexError:
                reason = None
        elif command == 'KY-ACTIVATE-CHECK':
            # 获取匹配的服务序列号
            serial = data[1]
            hwinfo = data[2]
        elif command == 'KY-DEACTIVATE-CHECK':
            # 检查取消配额
            serial = data[1]
            hwinfo = data[2]
        elif command == 'KY-CONFIG':
            pass
        else:
            app.logger.debug('unknown command: %s', command)
            cs.close()
            return

    except Exception:
        app.logger.debug('wrong format')
        cs.close()
        return

    if command == 'KY-ACTIVATE':
        try:
            key, user = kms_core.count_key(
                serial, regcode.replace('-', ''), addr[0], hwinfo,
                oem_serial, vendor, model, passive=True)
        except kms_core.ActivateError as e:
            response = ('KY-ACTIVATE-ERROR:' + str(e))
            app.logger.debug('response: %s', response)
            cs.send(response.encode('utf-8'))
        else:
            response = 'KY-ACTIVATE-KEY:{}:{}'.format(key, user)
            app.logger.debug('response: %s', response)
            cs.send(response.encode('utf-8'))
    elif command == 'KY-DEACTIVATE':
        try:
            status, _msg = kms_core.deactivate(serial, regcode, reason)
        except Exception:
            app.logger.exception('Failed deactivating')
            status = 'ERROR'
        response = 'KY-DEACTIVATE-STATUS:{}'.format(status)
        app.logger.debug('response: %s', response)
        cs.send(response.encode('utf-8'))
    elif command == 'KY-ACTIVATE-CHECK':
        # 获取匹配的服务序列号
        serial, regcode = kms_core.search_fit_serial(
            serial, hwinfo, addr[0])
        if not serial:
            app.logger.debug('no fit serial found')
        response = 'KY-ACTIVATE-SERIAL:{}'.format(serial)
        app.logger.debug('response: %s', response)
        cs.send(response.encode('utf-8'))
    elif command == 'KY-DEACTIVATE-CHECK':
        # 检查取消配额
        serial = data[1]
        hwinfo = data[2]
        quota = kms_core.get_deactivate_quota(
            serial, hwinfo)
        if quota is None:
            quota = -1
        response = 'KY-DEACTIVATE-QUOTA:{}'.format(quota)
        app.logger.debug('response: %s', response)
        cs.send(response.encode('utf-8'))
    elif command == 'KY-CONFIG':
        response = """KY-CONFIG-CONTENT:[Active_State]
activation_interval_sec=%d
retry_count=%d

[Inactive_State]
initial_interval_sec=%d
offset_value_sec=%d""" % (
            app_conf.getint('Active_State', 'activation_interval_sec'),
            app_conf.getint('Active_State', 'retry_count'),
            app_conf.getint('Inactive_State', 'initial_interval_sec'),
            app_conf.getint('Inactive_State', 'offset_value_sec'),
        )
        cs.send(response.encode('utf-8'))
    cs.close()
    close_db()


class Sync2Slave(threading.Thread):
    """
    同步数据到备端线程
    """
    def __init__(self):
        threading.Thread.__init__(self, name='sync')
        self.slave_http = app_conf.get('server', 'slave_url')
        self.stopped = False

    def run(self) -> None:
        # 启用数据同步队列
        kms_core.manager.sync_enabled = True
        while not self.stopped:
            try:
                # 同步queue中的数据
                if kms_core.sync_q.qsize() >= 1:
                    data = kms_core.sync_q.get()
                    req = urllib.request.Request(self.slave_http)
                    req.add_header('Content-Type', 'application/oct-stream; charset=utf-8')
                    # data = json.dumps(data)
                    # data = data.encode('utf-8')  # needs to be bytes
                    response = urllib.request.urlopen(req, data)
                    app.logger.debug(response.read())
                time.sleep(0.1)
            except Exception as e:
                import traceback
                app.logger.error(traceback.format_exc())

    def stop(self):
        app.logger.info('Stopping sync thread, please wait')
        self.stopped = True


class TcpServer(threading.Thread):  # 使用线程处理激活
    def __init__(self) -> None:
        threading.Thread.__init__(self, name='tcp')
        self.socket = socket(AF_INET, SOCK_STREAM)
        self.socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.epoll = select.epoll()

        self.stopped = False

    def run(self):
        s = self.socket
        s.settimeout(5)
        s.bind(load_bind_conf('tcp_bind', ('0.0.0.0', 7070)))

        s.listen(10)

        s.setblocking(False)

        epl = self.epoll
        epl.register(s.fileno(), select.EPOLLIN)

        fd_event_dict = dict()
        fd_client_dict = dict()

        while not self.stopped:
            try:
                fd_event_list = epl.poll(timeout=1)
                # s.listen(5)
                for fd, event in fd_event_list:
                    if fd == s.fileno():
                        new_socket, client_addr = s.accept()
                        # 监听到一个新的客户端连接，将new_socket也注册到epoll中
                        new_socket.settimeout(10)
                        app.logger.debug(
                            'got connected from {}'.format(client_addr))
                        epl.register(new_socket.fileno(), select.EPOLLIN)
                        # 并且将这个socket加入fd_event_dict字段，方便以后通过fd来获取套接字
                        fd_event_dict[new_socket.fileno()] = new_socket
                        fd_client_dict[new_socket.fileno()] = client_addr
                    # elif event ==  select.EPOLLIN:
                    else:
                        cs = fd_event_dict[fd]
                        addr = fd_client_dict[fd]
                        request = cs.recv(1024).decode('utf-8')
                        if request:
                            p = threading.Thread(
                                target=dealrequest, args=(request, cs, addr))
                            p.start()
                        else:
                            # 关闭fd对应的socket
                            fd_event_dict[fd].close()
                            # 从epoll中踢出已经断开的fd
                            epl.unregister(fd)
                            # 从字典中删除fd对应的记录
                            del fd_event_dict[fd]
                            del fd_client_dict[fd]
            except (OSError, timeout):
                pass
            except:
                app.logger.debug('listen error:{}'.format(
                    traceback.format_exc()))
                close_db()
        app.logger.info('tcp exit')

    def stop(self):
        app.logger.info('Stopping tcp, please wait')
        self.stopped = True
        self.socket.close()


class UdpServer(threading.Thread):
    def __init__(self) -> None:
        threading.Thread.__init__(self, name='udp')
        self.socket = socket(AF_INET, SOCK_DGRAM)
        self.socket.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
        self.socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.stopped = False
        self.epoll = select.epoll()

    def run(self):
        s = self.socket
        s.settimeout(5)
        s.bind(load_bind_conf('udp_bind', ('0.0.0.0', 7070)))
        s.setblocking(False)
        epoll = self.epoll
        epoll.register(s.fileno(), select.EPOLLIN)
        fd_to_socket = {s.fileno(): s, }

        while not self.stopped:
            try:
                events = epoll.poll(timeout=1)
                for fd, event in events:
                    socket = fd_to_socket[fd]
                    if event & select.EPOLLIN:
                        msg, addr = socket.recvfrom(1024)
                        # 如果有数据
                        if msg:
                            # 处理数据
                            request = msg.decode('utf-8')
                            app.logger.debug(
                                'recv from:{}, msg:{}'.format(addr, request))
                            # 只有master端才能处理激活、取消激活请求
                            if app_conf.get('server', 'mode') != 'master':
                                response = 'This server is slave, can not handler those operator!'
                                app.logger.debug(response)
                                socket.sendto(response.encode('utf-8'), addr)
                            try:
                                command, serial, hwinfo = request.split(':')
                                assert len(serial) >= 6
                            except:
                                app.logger.debug('wrong format')
                                continue
                            if command == 'KY-ACTIVATE':
                                # 获取匹配的服务序列号
                                serial, regcode = kms_core.search_fit_serial(
                                    serial, hwinfo, addr[0])
                                if not serial:
                                    app.logger.debug('no fit serial found')
                                    continue
                                response = 'KY-ACTIVATE-SERIAL:{}'.format(
                                    serial)
                            elif command == 'KY-DEACTIVATE-CHECK':
                                # 检查取消配额
                                quota = kms_core.get_deactivate_quota(
                                    serial, hwinfo)
                                if quota is None:
                                    continue
                                response = 'KY-DEACTIVATE-QUOTA:{}'.format(
                                    quota)
                            else:
                                app.logger.debug(
                                    'unknown command: %s', command)
                                continue
                            app.logger.debug('response: %s', response)
                            socket.sendto(response.encode('utf-8'), addr)
                        else:  # 如果没有数据，则表示客户端断开连接
                            # 从epoll中踢出已经断开的fd
                            epoll.unregister(fd)
                            fd_to_socket[fd].close
                            # 从字典中删除fd对应的记录
                            del fd_to_socket[fd]
                    elif event & select.EPOLLHUP:
                        # 在epoll中注销客户端的文件句柄
                        epoll.unregister(fd)
                        # 关闭客户端的文件句柄
                        fd_to_socket[fd].close()
                        # 在字典中删除与已关闭客户端相关的信息
                        del fd_to_socket[fd]
            except (OSError, timeout):
                pass
            except:
                app.logger.debug('listen error:{}'.format(
                    traceback.format_exc()))
                close_db()
        app.logger.info('udp exit')

    def stop(self):
        app.logger.info('Stopping udp, please wait')
        self.stopped = True
        self.epoll.unregister(self.socket.fileno())
        self.epoll.close()
        self.socket.close()


class Backup(threading.Thread):
    """
    数据备份线程
    """
    def __init__(self):
        threading.Thread.__init__(self, name='backup')
        self.backup_http = app_conf.get('backup', 'backup_url')
        self.file_path = app_conf.get('backup', 'file_path')
        try:
            self.backup_interval = app_conf.getint('backup', 'backup_interval')
        except ValueError:
            app.logger.debug('default backup interval: 600s')
            self.backup_interval = 600

        self.stopped = False
        # 为了及时响应命令行stop信号
        self.tick = 0

    def check_config(self):
        if not self.file_path and not self.backup_http:
            app.logger.info(
                'neither backup.file_path nor backup.backup_url is set, '
                'auto backup is disabled')
            self.stopped = True

    def local_backup(self, data):
        if not self.file_path:
            return

        if not os.path.exists(self.file_path):
            try:
                os.makedirs(self.file_path)
            except OSError:
                app.logger.exception('failed creating backup directory')
                return
        elif not os.path.isdir(self.file_path):
            app.logger.error(
                '%s should be a folder to store backup files',
                self.file_path)
            return

        if len(os.listdir(self.file_path)) > 0:
            # 这里有且仅有一个文件，先清理原有的同步数据文件
            os.system('rm %s/*' % self.file_path)
        # 将备份的数据存储到文件中
        filepath = os.path.join(self.file_path, str(datetime.datetime.now().strftime('%Y%m%d%H%M%S')))
        with open(filepath, 'wb') as f:
            f.write(data)

    def remote_backup(self, data):
        if not self.backup_http:
            return

        req = urllib.request.Request(self.backup_http + '/api/server/backup')
        req.add_header('Content-Type', 'application/oct-stream; charset=utf-8')
        response = urllib.request.urlopen(req, data)
        app.logger.debug(response.read().decode())

    def run(self) -> None:
        self.check_config()

        while not self.stopped:
            if self.tick < self.backup_interval:
                time.sleep(1)
                self.tick += 1
                continue
            else:
                self.tick = 0

            try:
                data = kms_core.get_backup_data()
                # 保存本地备份
                self.local_backup(data)
                # 发送到远端备份服务
                self.remote_backup(data)
            except Exception:
                app.logger.exception('Error on backup')

    def stop(self):
        app.logger.info('Stopping backup thread, please wait')
        self.stopped = True


def send_message(topic, body):
    json_info = json.dumps(body)
    result = client.publish(topic, json_info, qos=1,
                            properties=publish_properties)
    status = result[0]
    if status == 0:
        app.logger.info("Send to topic `%s`: `%s` ", topic, json_info)
    else:
        app.logger.warning("Failed to send message to topic %s", topic)
    return status


def on_connect(client, userdata, flags, reason, properties):
    if reason.value == 0:
        app.logger.info('Connected to MQTT Broker!')
        client.subscribe('kylin/register/request')  # 注册
        client.subscribe('kylin/activate/response')  # 激活响应
        client.subscribe('kylin/recycle/response')  # 回收响应
        client.subscribe('kylin/status/request')  # 状态上报
        client.subscribe('kylin/c_recycle/request')  # 客户端主动回收
        client.subscribe('kylin/c_activate/request')  # 客户端主动激活
    else:
        app.logger.debug("Failed to connect, reason %s\n", reason)


def on_message(client, userdata, msg):
    db = get_db()
    now = time.time()
    topic = msg.topic
    payload = json.loads(msg.payload.decode('utf-8').strip('\x00'))

    # get client ip
    client_ip = payload.get('ip', '')
    if isinstance(client_ip, list):
        client_ip = ', '.join(client_ip)

    app.logger.info("Msg from `%s`: `%s`", topic, payload)
    if topic == 'kylin/register/request':
        serial, regcode = kms_core.search_fit_serial(
            payload['ser_num'], payload['hd_code'], client_ip, mqtt_tag=True, mqtt_payload=payload)
        if not serial:
            app.logger.debug('no fit serial found')
            serial = None
            k_result = 1
        else:
            k_result = 0

            # 用于除去多余的扩展信息字段
            attr_dict = db(
                (db.client_extend_attr.f_regcode == regcode)
            ).select(db.client_extend_attr.f_key).as_list()
            attr_list = [attr_i['f_key'] for attr_i in attr_dict]

            # 存储客户端注册时的扩展信息字段
            for k, v in payload.items():
                if k[0:2] == 'X-':
                    key = k[2:]
                    if key in attr_list:
                        attr_list.remove(key)
                    attr_row = db(
                        (db.client_extend_attr.f_regcode == regcode) &
                        (db.client_extend_attr.f_key == key)
                    ).select().first()
                    if attr_row:
                        attr_row.update_record(
                            f_value=v,
                            f_change=now
                        )
                    else:
                        db.client_extend_attr.insert(
                            f_regcode=regcode,
                            f_key=key,
                            f_value=v,
                            f_change=now
                        )

            # 删除数据库中多余的扩展信息字段
            if len(attr_list) != 0:
                db(
                    (db.client_extend_attr.f_regcode == regcode) &
                    db.client_extend_attr.f_key.belongs(attr_list)
                ).delete()

        response_dict = {'version': protocol_version,
                         'ser_num': serial,
                         'hd_code': payload['hd_code'],
                         'ret': k_result}
        response = 'kylin/register/response'
        send_message(response, response_dict)
    elif topic == 'kylin/status/request':
        try:
            result_update = kms_core.update_activation(payload, topic)
        except:
            app.logger.exception('Failed update client')
            result_update = 1

        response_dict = {'version': protocol_version,
                         'ret': result_update}
        sn_number = str(payload['SN'])
        sn_number.replace('-', '')
        response = 'kylin/status/response/'+sn_number
        send_message(response, response_dict)

    elif topic == 'kylin/c_activate/request':
        c_result = 0
        key = None
        user = None
        # 服务期限
        term = ''
        try:
            # 获取匹配的服务序列号
            serial, regcode = kms_core.search_fit_serial(
                payload['ser_num'], payload['hd_code'], client_ip, mqtt_tag=True, mqtt_payload=payload)
            if not serial:
                app.logger.debug('no fit serial found')
                c_result = 1
            else:
                client_regcode = payload['SN']
                if not client_regcode or len(client_regcode.replace('-', '')) != 20:
                    client_regcode = regcode
                key, user = kms_core.count_key(
                    serial, client_regcode, client_ip,
                    None, None, None, None, mqtt_tag=True, passive=True)
                # service term
                term = kms_core.get_serial_term(serial)
        except kms_core.ActivateError as e:
            app.logger.error('Failed activate: %s', e)
            c_result = 1
            serial = payload['ser_num']
        except:
            app.logger.exception('Failed handle activate request')
            c_result = 1
            serial = payload['ser_num']
        response = 'kylin/c_activate/response'

        response_dict = {'version': protocol_version,
                         'ser_num': serial,
                         'hd_code': payload['hd_code'],
                         'act_code': key,
                         'customer': user,
                         'term': term,
                         'ret': c_result}
        send_message(response, response_dict)
    elif topic == 'kylin/c_recycle/request':
        r_result = 0
        try:
            status_dea, msg = kms_core.deactivate(None, payload['SN'],
                                             reason='客户端主动取消',
                                             act_code=payload['act_code'],
                                             local=True,
                                             confirm=True,
                                             force=False)
            if status_dea == 'SUCCESS':
                pass
            else:
                r_result = 1
                app.logger.warning(
                    'Failed deactivating, result: %s', msg)

        except Exception:
            app.logger.exception('Failed deactivating')
            r_result = 1
        response = 'kylin/c_recycle/response/'+payload['SN']
        response_dict = {'version': protocol_version,
                         'SN': payload['SN'],
                         'act_code': payload['act_code'],
                         'ret': r_result}
        send_message(response, response_dict)
    elif topic == 'kylin/activate/response':
        try:
            kms_core.update_activation(payload, topic)
        except:
            app.logger.exception('Failed handle activate response')
    elif topic == 'kylin/recycle/response':
        try:
            kms_core.update_activation(payload, topic)
        except:
            app.logger.exception('Failed handle recycle response')
    close_db()


def timing(data_list):
    time.sleep(60)
    db = get_db()
    untreated_list = []
    for data_t in data_list:
        serial = data_t.get('serial')
        clientkey = data_t.get('clientKey')
        row = db((db.activation.f_serial == serial) &
                 (db.activation.f_regcode == kms_core.code_add_hyphen(clientkey))
                 ).select().first()
        if not row:
            break
        row_dict = row.as_dict()
        if row_dict['f_batchactivate_tag'] == 2:
            untreated_list.append(data_t)

    for data_x in untreated_list:
        serial_x = data_x.get('serial')
        clientkey_x = data_x.get('clientKey')
        # 超时的默认设置为激活失败
        update_info = {'f_batchactivate_tag': 3, 'f_status': 2}
        if data_x.get('old_status') == 0:
            # 延期失败的恢复之前的激活码和服务期限，不回收
            update_info['f_key'] = data_x['old_key']
            update_info['f_term'] = data_x['old_term']
            update_info['f_status'] = 0
        else:
            # 回收激活码
            status_dea, msg = kms_core.deactivate(
                serial_x, clientkey_x, reason='批量激活超时', local=True, force=False)
            if status_dea != 'SUCCESS':
                app.logger.warning(
                    'Fail to deactivate %s: %s', clientkey_x, msg)
            else:
                app.logger.info(
                    'Succeed to deactivate %s on timeout', clientkey_x)
        db(db.activation.f_regcode == kms_core.code_add_hyphen(
            clientkey_x)).update(**update_info)
    close_db()


def timing_cancel(data_list):
    time.sleep(60)
    db = get_db()
    untreated_list = []
    for data_t in data_list:
        serial = data_t.get('serial')
        clientkey = data_t.get('clientKey')
        row = db((db.activation.f_serial == serial) &
                 (db.activation.f_regcode == kms_core.code_add_hyphen(clientkey))
                 ).select().first()
        if not row:
            break
        row_dict = row.as_dict()
        if row_dict['f_batchdeactivate_tag'] == 2:
            untreated_list.append(data_t)

    for data_x in untreated_list:
        clientkey_x = data_x.get('clientKey')
        # 超时的默认设置f_batchdeactivate_tag为3
        update_info = {'f_batchdeactivate_tag': 3}
        db(db.activation.f_regcode == kms_core.code_add_hyphen(
            clientkey_x)).update(**update_info)
    close_db()


@app.route('/api/server/batchactivate', methods=['POST'])
@requires_local
def batchactivate():
    db = get_db()
    data = request.stream.read()
    try:
        data_list = json.loads(data.decode('utf-8'))
    except json.JSONDecodeError:
        abort(400)
    ret_code = 0
    msg = 'Success'
    batchactivate_tag = False
    serial_num = {}
    to_activate = []
    for data_j in data_list:
        serial = data_j.get('serial')
        clientkey = data_j.get('clientKey')
        row = db((db.activation.f_serial == serial) &
                 (db.activation.f_regcode == kms_core.code_add_hyphen(clientkey))
                 ).select().first()
        if serial in serial_num.keys():
            serial_num[serial] = serial_num[serial]+1
        else:
            serial_num[serial] = 1
        if row:
            data_j['old_status'] = row.f_status
            if row.f_status == 0:
                # 将该条记录的f_batchactivate_tag设置为5，表示未在mqtt中发送激活信息
                row.update_record(f_batchactivate_tag=5)
                db.commit()
                # 不占用次数
                serial_num[serial] = serial_num[serial] - 1
                # 已激活的允许延期
                license = kms_core.manager.get_license(
                    row.f_activate_by or serial)
                if license and license.activate_term(row.f_regcode) > row.f_term:
                    # 可以再次激活以延期
                    data_j['old_term'] = row.f_term
                    data_j['old_key'] = row.f_key
                else:
                    # 跳过已激活的
                    app.logger.debug(
                        'Client %s is already activated, skip', clientkey)
                    continue
        batchactivate_tag = True
        to_activate.append(data_j)
    if batchactivate_tag:
        for key, value in serial_num.items():
            if not value:
                continue
            serial_row = db((db.serial.f_serial == key)
                            ).select().first()
            if serial_row:
                if serial_row.f_mode & kms_core.ActivateMode.unlimited:
                    # 不限量模式不用检查剩余次数
                    pass
                else:
                    serial_left = serial_row.f_left
                    if value and serial_left < value:
                        batchactivate_tag = False
                        ret_code = 1
                        msg = 'The rest amount of %s is not enough' % str(key)
                        break
            else:
                batchactivate_tag = False
                ret_code = 1
                msg = 'Serial %s not found' % str(key)
                break
    if batchactivate_tag:
        activated = []
        for data_i in to_activate:
            serial = data_i.get('serial')
            clientkey = data_i.get('clientKey')
            try:
                key, user = kms_core.count_key(
                    serial, clientkey, None, None, None, None, None, mqtt_tag=True)
                response = 'kylin/activate/request/'+str(clientkey)
                response_dict = {'version': protocol_version,
                                 'act_code': key,
                                 'customer': user,
                                 'term': kms_core.get_serial_term(serial)
                                 }
                msg_status = send_message(response, response_dict)

                if msg_status != 0:
                    # deactivate and reset status
                    update_info = {'f_status': 2}
                    update_info['f_batchactivate_tag'] = 4
                    if data_i.get('old_status') == 0:
                        # 延期失败的恢复之前的激活码和服务期限，不回收
                        update_info['f_key'] = data_i['old_key']
                        update_info['f_term'] = data_i['old_term']
                        update_info['f_status'] = 0
                    else:
                        # 回收激活码
                        status_dea, msg = kms_core.deactivate(
                            serial, clientkey, reason='批量激活失败', local=True, force=False)
                        if status_dea != 'SUCCESS':
                            app.logger.warning(
                                'Fail to deactivate %s: %s', clientkey, msg)
                        else:
                            app.logger.info(
                                'Succeed to deactivate %s on communitcation failure', clientkey)
                    db((db.activation.f_serial == serial) &
                       (db.activation.f_regcode ==
                        kms_core.code_add_hyphen(clientkey))
                       ).update(**update_info)
                    ret_code = 1
                    msg = 'Fail to activate %s' % clientkey
                    app.logger.error('stop acitvate on communitcation failure')
                    # 发送激活信息失败时，继续处理其他客户端
                    continue
                # 等待客户端响应
                db((db.activation.f_serial == serial) &
                   (db.activation.f_regcode == kms_core.code_add_hyphen(clientkey))
                   ).update(f_batchactivate_tag=2)
                activated.append(data_i)
            except Exception as e:
                if isinstance(e, kms_core.ActivateError):
                    app.logger.error('batch activate error: %s', e)
                else:
                    app.logger.exception(
                        'Failed handle batch activate request')
                ret_code = 1
                msg = 'Fail to activate %s' % clientkey
                break

        timeThread_activate = threading.Thread(
            target=timing, args=(activated,))
        timeThread_activate.start()
    app.logger.debug('batch activate response: %s', msg)
    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/batchcancel', methods=['POST'])
@requires_local
def batchcancel():
    db = get_db()
    data = request.stream.read()
    try:
        data_list = json.loads(data.decode('utf-8'))
    except json.JSONDecodeError:
        abort(400)
    ret_code = 0
    msg = 'Success'
    batchcancel_tag = True
    to_cancel = []
    serial_cancels = {}
    for data_j in data_list:
        serial = data_j.get('serial')
        clientkey = data_j.get('clientKey')
        row = db((db.activation.f_serial == serial) &
                 (db.activation.f_regcode == kms_core.code_add_hyphen(clientkey))
                 ).select().first()
        if row:
            # check serial
            if row.f_status != 0:
                # 将该条记录的f_batchdeactivate_tag设置为5，表示未在mqtt中发送取消激活信息
                row.update_record(f_batchdeactivate_tag=5)
                db.commit()
                app.logger.debug('Client %s is not activated, skip', clientkey)
                continue
            # check license
            license = kms_core.manager.get_license(row.f_activate_by or serial)
            if not license:
                batchcancel_tag = False
                ret_code = 1
                msg = 'Serial %s not found' % serial
                app.logger.warning(msg)
                break
            data_j['f_key'] = row.f_key
            to_cancel.append(data_j)
            if license.serial in serial_cancels:
                serial_cancels[license.serial] += 1
            else:
                serial_cancels[license.serial] = 1
            if not license.check_deactivate_quota(serial_cancels[license.serial]):
                batchcancel_tag = False
                ret_code = 1
                msg = 'Serial %s is out of deactivate quota' % serial
                app.logger.warning(msg)
                break
        else:
            batchcancel_tag = False
            ret_code = 1
            msg = 'Client %s not found' % clientkey
            break

    if batchcancel_tag:
        canceled = []
        for data_i in to_cancel:
            serial = data_i.get('serial')
            clientkey = data_i.get('clientKey')

            request_dict = {'version': protocol_version,
                            'act_code': data_i['f_key']
                            }

            msg_status = send_message(
                'kylin/recycle/request/'+str(clientkey), request_dict)
            if msg_status != 0:
                ret_code = 1
                msg = 'Fail to send deactivate message for %s' % clientkey
                app.logger.error('stop acitvate on communitcation failure')
                # 发送取消激活信息失败时，将f_batchdeactivate_tag设置为状态4，继续处理其他客户端
                db((db.activation.f_serial == serial) &
                   (db.activation.f_regcode == kms_core.code_add_hyphen(clientkey))
                   ).update(f_batchdeactivate_tag=4)
                db.commit()
                continue
            # 等待客户端取消激活的响应
            db((db.activation.f_serial == serial) &
               (db.activation.f_regcode == kms_core.code_add_hyphen(clientkey))
               ).update(f_batchdeactivate_tag=2)
            canceled.append(data_i)
        timeThread_deactivate = threading.Thread(
            target=timing_cancel, args=(canceled,))
        timeThread_deactivate.start()
    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/backup', methods=['POST'])
def backup():
    data = request.stream.read()
    result = kms_core.update_local_data(data)
    if result:
        ret_code = 0
        msg = 'Success'
        app.logger.info('Succeed to sync master data')
    else:
        ret_code = -1
        msg = 'Failure'
        app.logger.info('Failure to sync master data')
    return jsonify({'return_code': ret_code, 'return_msg': msg})


@app.route('/api/server/batchactivatestatus', methods=['POST'])
@requires_local
def batchactivatestatus():
    db = get_db()
    data = request.stream.read()
    try:
        data_list = json.loads(data.decode('utf-8'))
    except json.JSONDecodeError:
        abort(400)
    ret_code = 0
    msg = 'Success'
    pending_activate = []
    activated = []
    activate_timeout = []
    activate_faild = []

    for data_j in data_list:
        row = db(
                 (db.activation.f_regcode == kms_core.code_add_hyphen(data_j))
                 ).select().first()
        if row:
            # f_status为0表示已激活状态，也是默认状态
            if row.f_status == 0:
                # 当f_batchactivate_tag为1：客户端批量激活已经处理，说明激活成功
                if row.f_batchactivate_tag == 1:
                    activated.append(row.f_regcode)
                # 当f_batchactivate_tag为2：客户端批量激活未处理，说明处于待激活
                elif row.f_batchactivate_tag == 2:
                    pending_activate.append(row.f_regcode)
                # 当f_batchactivate_tag为3：客户端响应激活超时，但该之前激活状态old_status为0
                elif row.f_batchactivate_tag == 3:
                    activated.append(row.f_regcode)
                # 当f_batchactivate_tag为4：发送激活请求信息失败，但该之前激活状态old_status为0
                elif row.f_batchactivate_tag == 4:
                    activated.append(row.f_regcode)
                # 当f_batchactivate_tag为5：因该条激活记录已被激活未发送激活请求信息
                elif row.f_batchactivate_tag == 5:
                    activated.append(row.f_regcode)
                # f_batchactivate_tag的其他情况，视为该请求处于未处理阶段
                else:
                    pending_activate.append(row.f_regcode)
            # f_status为2表示激活失败状态
            elif row.f_status == 2:
                # 当f_batchactivate_tag为3，说明客户端响应激活超时
                if row.f_batchactivate_tag == 3:
                    activate_timeout.append(row.f_regcode)
                else:
                    activate_faild.append(row.f_regcode)
            # f_status为1表示待激活状态
            elif row.f_status == 1:
                # f_batchactivate_tag为1和f_status为1说明客户端返回激活失败的信息
                if row.f_batchactivate_tag == 1:
                    activate_faild.append(row.f_regcode)
        else:
            ret_code = 1
            msg = 'Client %s not found' % kms_core.code_add_hyphen(data_j)
            return jsonify({'return_code': ret_code, 'return_msg': msg})

    return jsonify({'pending_activate': pending_activate,
                    'activated': activated,
                    'activate_timeout': activate_timeout,
                    'activate_faild': activate_faild,
                    'return_code': ret_code,
                    'return_msg': msg})


@app.route('/api/server/batchcancelstatus', methods=['POST'])
@requires_local
def batchcancelstatus():
    db = get_db()
    data = request.stream.read()
    try:
        data_list = json.loads(data.decode('utf-8'))
    except json.JSONDecodeError:
        abort(400)
    ret_code = 0
    msg = 'Success'
    pending_cancel = []
    cancelled = []
    cancel_timeout = []
    cancel_faild = []

    for data_j in data_list:
        row = db(
                 (db.activation.f_regcode == kms_core.code_add_hyphen(data_j))
                 ).select().first()
        if row:
            # f_status为0表示已激活状态
            if row.f_status == 0:
                # 当f_batchdeactivate_tag为0：0为初始状态，表明未对该激活记录进行处理
                if row.f_batchdeactivate_tag == 0:
                    pending_cancel.append(row.f_regcode)
                # 当f_batchdeactivate_tag为1：客户端响应了批量取消激活的请求，但f_status状态未改变表明客户端取消激活失败
                elif row.f_batchdeactivate_tag == 1:
                    cancel_faild.append(row.f_regcode)
                # 当f_batchdeactivate_tag为2：等待客户端响应批量取消激活的请求
                elif row.f_batchdeactivate_tag == 2:
                    pending_cancel.append(row.f_regcode)
                # 当f_batchdeactivate_tag为3：客户端响应批量取消激活请求超时
                elif row.f_batchdeactivate_tag == 3:
                    cancel_timeout.append(row.f_regcode)
                # 当f_batchdeactivate_tag为4：客户端发送批量取消激活信息失败，导致取消激活失败
                elif row.f_batchdeactivate_tag == 4:
                    cancel_faild.append(row.f_regcode)
            # f_status为4表示激活已取消状态
            elif row.f_status == 4:
                # 当f_batchdeactivate_tag为1：客户端响应了批量取消激活的请求，f_status状态变为4表明客户端取消激活成功
                if row.f_batchdeactivate_tag == 1:
                    cancelled.append(row.f_regcode)
                # 当f_batchdeactivate_tag为5：未向客户端发送取消激活的请求
                elif row.f_batchdeactivate_tag == 5:
                    cancelled.append(row.f_regcode)
            else:
                # 当f_batchdeactivate_tag为5：未向客户端发送取消激活的请求
                if row.f_batchdeactivate_tag == 5:
                    cancelled.append(row.f_regcode)
        else:
            ret_code = 1
            msg = 'Client %s not found' % kms_core.code_add_hyphen(data_j)
            return jsonify({'return_code': ret_code, 'return_msg': msg})

    return jsonify({'pending_cancel': pending_cancel,
                    'cancelled': cancelled,
                    'cancel_timeout': cancel_timeout,
                    'cancel_faild': cancel_faild,
                    'return_code': ret_code,
                    'return_msg': msg})


def on_disconnect(client, userdata, rc):
    if rc != 0:
        app.logger.error("Unexpected disconnection %s" % rc)
        app.logger.info('try reconnect')
        client.reconnect()


SERVER_ON = True
def mqttCommunication():
    try:
        broker_address = app_conf["mqtt"]["broker"]
        broker_ip = broker_address.split(':')[0]
        broker_port = int(broker_address.split(':')[1])
    except KeyError:
        broker_ip = '127.0.0.1'
        broker_port = 1883
    client.on_connect = on_connect
    client.reconnect_delay_set(max_delay=10)
    app.logger.debug('connecting mqtt broker %s', broker_ip)
    while SERVER_ON:
        try:
            client.connect(broker_ip, broker_port)
            break
        except:
            time.sleep(5)

    client.on_message = on_message
    client.on_disconnect = on_disconnect
    client.loop_forever()


class ExpireClientThread(threading.Thread):
    stopped = False

    def run(self):
        while 1:
            try:
                app_conf.reload()
                interval = app_conf.getint('Active_State', 'activation_interval_sec')
                retry_count = app_conf.getint('Active_State', 'retry_count')
                kms_core.expire_clients(cloud_expire=interval * retry_count)

                for _i in range(interval):
                    if self.stopped:
                        return
                    time.sleep(1)

            except Exception:
                app.logger.exception('Failed running expire_clients')

    def stop(self):
        self.stopped = True


if __name__ == "__main__":
    kms_core.manager.passive = app_conf.getboolean('server', 'passive_mode')

    t2 = UdpServer()
    t2.start()
    t3 = TcpServer()
    t3.start()
    mqttThread = threading.Thread(target=mqttCommunication, name='msg-client')
    mqttThread.start()
    backup_obj = Backup()
    backup_obj.start()
    if app_conf.getboolean('server', 'master_slave'):
        sync = Sync2Slave()
        sync.start()
    expire_clients_thread = ExpireClientThread(name='expire-clients')
    expire_clients_thread.start()
    listen_http() # Ctrl + C to stop
    # stopped
    SERVER_ON = False
    t2.stop()
    t3.stop()
    expire_clients_thread.stop()
    backup_obj.stop()
    if app_conf.getboolean('server', 'master_slave'):
        sync.stop()

    client.on_disconnect = None
    client.disconnect()
    client.loop_stop()
